/*
 * Copyright (c) 2004, Nate Nielsen
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 * 
 *     * Redistributions of source code must retain the above 
 *       copyright notice, this list of conditions and the 
 *       following disclaimer.
 *     * Redistributions in binary form must reproduce the 
 *       above copyright notice, this list of conditions and 
 *       the following disclaimer in the documentation and/or 
 *       other materials provided with the distribution.
 *     * The names of contributors to this software may not be 
 *       used to endorse or promote products derived from this 
 *       software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
 * DAMAGE.
 * 
 *
 * CONTRIBUTORS
 *  Nate Nielsen <nielsen@memberwebs.com>
 *
 */ 

#include "usuals.h"
#include "rtfanalyser.h"
#include "rtffixups.h"
#include "domhelpers.h"
#include "tags.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

RtfParser::RtfParser(const RtfParserOptions& options)
{
	m_document = NULL;
	memcpy(&m_options, &options, sizeof(options));
    
    for(int i = 0; i < AUTOCOUNT_MAX; i++)
        m_autocount[i] = 1;
}

RtfParser::~RtfParser()
{
	clear();

	if(m_impl != NULL)
		m_impl.release();
}

void RtfParser::clear()
{
	if(m_document != NULL)
	{
		try
		{
			m_document.release();
		}
		catch(...) { }

		m_document = NULL;
	}
	LevelHandler::clear();
}

void RtfParser::startDocument(RtfReader* reader)
{
	LevelHandler::startDocument(reader);

	// Create a new document
	m_document = m_impl.createDocument("", kElDoc, DOM::DocumentType());

	// TODO: Throw error if document is null
	ASSERT(m_document != NULL);

	ASSERT(m_curLevel != NULL);
	m_curLevel->setElement(m_document.getDocumentElement(), true);

	// Set the attributes on the top level
	setAnalyser(AnalyserPtr(new Root));
	setDestination(DestinationPtr(new Content));
	getTextFormatting().resetPara();
	getTextFormatting().resetText();
}

void RtfParser::endDocument()
{
	LevelHandler::endDocument();

	// Cleanup the tree
	RtfFixups::removeDuplicates(m_document);
	RtfFixups::consolidateStartTags(m_document);
    RtfFixups::consolidateEndTags(m_document);
	RtfFixups::breakTables(m_document);
	RtfFixups::breakTags(m_document, kElTable, kElRow);
	RtfFixups::breakTags(m_document, kElRow, kElCell);
	RtfFixups::wrapTags(m_document, kElCell, kElDest);
	RtfFixups::breakBlocks(m_document);
	RtfFixups::breakLists(m_document);
	RtfFixups::fixLists(m_document);
	RtfFixups::fixStyles(m_document);
	RtfFixups::fixBlocks(m_document);
	RtfFixups::removeTags(m_document);
	RtfFixups::breakBreak(m_document, kElDoc, kElPage);
	RtfFixups::breakBreak(m_document, kElDoc, kElSect); 
    RtfFixups::removeDuplicates(m_document);
	return;
}





// -----------------------------------------------------------------------
//  Helper functions

DOM::Element RtfParser::createElement(const string& name)
{
	ASSERT(name.length() > 0);
	return m_document.createElement(name);

	// TODO: Throw exception here if necessary
}

void RtfParser::replaceElement(const DOM::Element& element)
{
	ASSERT(m_curLevel != NULL);
	m_curLevel->setElement(element, true);
}

void RtfParser::pushElement(const DOM::Element& element)
{
	ASSERT(m_curLevel != NULL);
	getElement().appendChild(element);
	m_curLevel->setElement(element);
}

DOM::Element RtfParser::popElement()
{
	DOM::Element element = getElement();
	ASSERT(m_curLevel != NULL);

	DOM::Node parent = element.getParentNode();
	ASSERT(parent.getNodeType() == DOM::Node::ELEMENT_NODE);

	// Set it deep so it replaces the current element
	m_curLevel->setElement((DOM::Element&)parent, true);
	return element;
}

void RtfParser::setAttribute(const string& name, const wstring& value, DOM::Element el)
{
	ASSERT(name.length() > 0);
	if(el == NULL)
		el = getElement();
	el.setAttribute(name, value);
}

void RtfParser::setAttribute(const string& name, int value, DOM::Element el)
{
    ASSERT(name.length() > 0);
    if(el == NULL)
        el = getElement();
    el.setAttribute(name, formatInt(value));
}

void RtfParser::setDestination(DestinationPtr dest)
{
	ASSERT(m_curLevel);

	m_curLevel->setDestination(dest);
	dest->m_parser = this;
	dest->initialize();
}

DestinationPtr RtfParser::replaceDestination(DestinationPtr dest)
{
	ASSERT(m_curLevel);
	
	DestinationPtr old = m_curLevel->getDestination();
	m_curLevel->setDestination(dest, true);
	dest->m_parser = this;
	dest->initialize();

	return old;
}


void RtfParser::setAnalyser(AnalyserPtr analy)
{
	ASSERT(m_curLevel);
	ASSERT(analy != NULL);

	analy->m_parser = this;
	m_curLevel->setAnalyser(analy);
	analy->initialize();
}

AnalyserPtr RtfParser::getAnalyser()
{
	ASSERT(m_curLevel);
	return m_curLevel->getAnalyser();
}

DestinationPtr RtfParser::getDestination()
{
	ASSERT(m_curLevel);
	return m_curLevel->getDestination();
}

RtfFormatting& RtfParser::getTextFormatting()
{
	ASSERT(m_curLevel);
	return m_curLevel->getFormatting();
}

int RtfParser::getAutoCount(int type)
{
    ASSERT(type < AUTOCOUNT_MAX);
    return m_autocount[type];
}

void RtfParser::incrementAutoCount(int type)
{
    ASSERT(type < AUTOCOUNT_MAX);
    m_autocount[type]++;
}

// ---------------------------------------------------------------------------------
//  Pass this stuff on through to the appropriate analysers etc...

void RtfParser::charData(wstring data)
{
	ASSERT(m_curLevel != NULL);
	DestinationPtr destination = m_curLevel->getDestination();
	if(destination)
	{
		destination->charData(data);
	}
	else
	{
		destination = DestinationPtr(new Content);
		setDestination(destination);
	}

}

void RtfParser::controlWord(const string& cw, int flags, int param)
{
	ASSERT(m_curLevel != NULL);
	AnalyserPtr analyser = m_curLevel->getAnalyser();
	if(analyser)
		analyser->controlWord(cw, flags, param);
}

void RtfParser::groupStart()
{
	LevelHandler::groupStart();

	ASSERT(m_curLevel != NULL);
	AnalyserPtr analyser = m_curLevel->getAnalyser();
	if(analyser)
		analyser->groupStart();
}

void RtfParser::groupEnd()
{
	ASSERT(m_curLevel != NULL);
    bool done = true;
    
	LevelHandler::groupEnd();

	AnalyserPtr analyser = m_curLevel->getAnalyser();
    if(analyser)
		analyser->groupEnd();
}

#define ON_INITIALIZE(cls) \
	void RtfParser::cls::initialize()
#define ON_CONTROLWORD(cls) \
	void RtfParser::cls::controlWord(const string& cw, int flags, int param)
#define ON_CHARDATA(cls) \
	void RtfParser::cls::charData(wstring data)
#define ON_GROUPSTART(cls) \
	void RtfParser::cls::groupStart()
#define ON_GROUPEND(cls) \
	void RtfParser::cls::groupEnd()
#define ON_DONE(cls) \
	void RtfParser::cls::done()
#define AN_ELEMENT(name) \
	m_parser->pushElement(m_parser->createElement(name))
#define AN_POP_ELEMENT() \
	m_parser->popElement()
#define AN_ATTRIBUTE(name, value) \
	m_parser->setAttribute(name, value)
#define AN_DESTINATION_ATTR(name) \
	m_parser->setDestination(new Attribute(name))
#define AN_DESTINATION(cls) \
	m_parser->setDestination(new cls)
#define AN_ANALYSER(cls) \
	m_parser->setAnalyser(AnalyserPtr(new cls))
#define AN_SET_ANALYSER(cls) \
	m_parser->setAnalyser(AnalyserPtr(cls))
#define HAS_PARAM (flags & kHasParam)
#define DEFAULT_CONTROLWORD processDefault(cw, flags, param)
#define DUMMY	1 == 1
#define NUM_ATTR(n) m_parser->formatInt(n)

bool RtfParser::ParseAnalyser::processDefault(const string& cw, int flags, int param)
{
	if(cw == "upr")
	{
		AnalyserPtr analy = m_parser->getAnalyser();
		ASSERT(analy != NULL);
		AN_SET_ANALYSER(new Upr(analy));
		return true;
	}

	return false;
}

void RtfParser::ParseAnalyser::applyParaFormatting(RtfFormatting* format, 
												   DOM::Element& el)
{
	if(format == NULL)
		format = &(m_parser->getTextFormatting());

	wstring fix = kValPara;

	int list = format->paraList();
	if(list != -1)
	{
		el.setAttribute(kAtList, NUM_ATTR(list));
	}
	else
	{
		el.removeAttribute(kAtList);
	}

	if(format->paraInTable())
		el.setAttribute(kAtCell, L"1");
	else
		el.removeAttribute(kAtCell);

	int style = format->paraStyle();
	if(style != -1)
		el.setAttribute(kElStyle, NUM_ATTR(style));
	else
		el.removeAttribute(kElStyle);

	el.setAttribute(kAtFix, fix);
}

DOM::Element RtfParser::ParseAnalyser::getCurrentBlock()
{
	DOM::Node node = m_parser->getElement();

	if(node.hasChildNodes())
		node = node.getLastChild();

	return DOMHelpers::getPriorElement(node, kElBlock);

}

bool RtfParser::ParseAnalyser::processTextContent(const string& cw, int flags, int param)
{
	DOM::Element el;
	bool process = false;

	RtfFormatting& format = m_parser->getTextFormatting();

	if(cw == "par")
	{
		el = getCurrentBlock();
		if(el != NULL)
			applyParaFormatting(&format, el);

		el = m_parser->createElement(kElBlock);
		applyParaFormatting(&format, el);
	}

	else if(cw == "intbl")
		format.paraSetTable(true);

	else if(cw == "cell")
	{
		el = getCurrentBlock();
		if(el != NULL)
			applyParaFormatting(&format, el);

		el = m_parser->createElement(kElCell);
		m_parser->pushElement(el);
		m_parser->popElement();
		el = m_parser->createElement(kElBlock);
		applyParaFormatting(&format, el);
	}

	else if(cw == "trowd")
		el = m_parser->createElement(kElRow);

	else if(cw == "tab")
		el = m_parser->createElement(kElTab);

	else if(cw == "sect")
		el = m_parser->createElement(kElSect);

	else if(cw == "page")
		el = m_parser->createElement(kElPage);

	else if(cw == "s" && HAS_PARAM)
		format.paraSetStyle(param);

	else if(cw == "line")
		el = m_parser->createElement(kElLine);

	else if(cw == "header")
		AN_ANALYSER(Skip);
	else if(cw == "footer")
		AN_ANALYSER(Skip);
	else if(cw == "bkmkstart")
		AN_ANALYSER(Skip);
	else if(cw == "listtext")
		AN_ANALYSER(Skip);
    
	else if(cw == "ls" && HAS_PARAM)
		format.paraSetList(param);

	if(el != NULL)
	{
		// This ensures that our content destination is open and ready
        DestinationPtr dest = m_parser->getDestination();
		ASSERT(dest != NULL);
		dest->charData(kValNull);

		m_parser->pushElement(el);
		m_parser->popElement();
	}

	return (el != NULL) || process;

	/* TODO: cell, row, intbl, cellx, trowd*/
}

bool RtfParser::ParseAnalyser::processTextFormatting(const string& cw, int flags, 
												  int param, RtfFormatting& format)
{
	bool on = true;
	if(flags & HAS_PARAM && param == 0)
		on = false;

	if(cw == "pard")
	{
		format.resetPara();
//		applyParaFormatting();
	}
	else if(cw == "plain")
		format.resetText();
	else if(cw == "b")
		format.textSetBold(on);
	else if(cw == "i")
		format.textSetItalic(on);
	else if(cw == "v")
		format.textSetHidden(on);
	else if(cw == "ul")
		format.textSetUnderline(on);
	else if(cw == "cf" && HAS_PARAM)
		format.textSetColor(param);
    else if(cw == "super")
        format.textSetSuScript(RtfFormatting::SUPERSCRIPT);
    else if(cw == "sub")
        format.textSetSuScript(RtfFormatting::SUBSCRIPT);
	else 
		return false;

	return true;
}

bool RtfParser::ParseAnalyser::processTextFormatting(const string& cw, int flags, int param)
{
	return processTextFormatting(cw, flags, param, m_parser->getTextFormatting());
}

bool RtfParser::ParseAnalyser::processTextAutoContent(const string& cw, int flags, int param)
{
    DestinationPtr dest = m_parser->getDestination();
    ASSERT(dest != NULL);
    dest->charData(kValNull);
  
    // Auto generated content
    if(cw == "chftn")
    {
        int ac = m_parser->getAutoCount(AUTOCOUNT_FOOTNOTE);
        
        AN_ELEMENT(kElRef);
        AN_ATTRIBUTE(kAtType, kValFootNote);
        AN_ATTRIBUTE(kAtTo, ac); 
        dest->charData(formatInt(ac));
        AN_POP_ELEMENT();
        return true;
    }      
    
    return false;
}

ON_INITIALIZE(Skip)
	{ AN_DESTINATION(Null); }
ON_GROUPSTART(Skip)
	{ AN_ANALYSER(Skip); }


RtfParser::Upr::Upr(AnalyserPtr prv)
{
	ASSERT(prv);
	prev = prv;
}
ON_GROUPSTART(Upr)
	{ AN_ANALYSER(Skip); }
ON_GROUPEND(Upr)
{ 
	ASSERT(prev);
	m_parser->setAnalyser(prev); 
	prev = NULL;
}


ON_INITIALIZE(Stylesheet)
{ 
	AN_ELEMENT(kElStylesheet); 
}
ON_GROUPSTART(Stylesheet)
{ 
	AN_ANALYSER(Style); 
	AN_DESTINATION(Null);
}



ON_INITIALIZE(Style)
{
	// Were not sure if this element is really something
	// so we can't always create
	haveStyle = false;
}
ON_CONTROLWORD(Style)
{
	// Get the style id 
	if(flags & kAsterisk)
	{
		AN_ANALYSER(Skip);
		return;
	}

	if(!haveStyle)
	{
		AN_ELEMENT(kElStyle);
		AN_DESTINATION_ATTR(kAtName);
		haveStyle = true;
	}

	if(cw == "s" && flags & kHasParam)
	{
		AN_ATTRIBUTE(kAtId, NUM_ATTR(param));
	}

	// Otherwise get as much formatting out of the tag as possible
	else if(processTextFormatting(cw, flags, param))
		DUMMY;

	else
		DEFAULT_CONTROLWORD;
}
ON_GROUPSTART(Style)
{ 
	AN_ANALYSER(Skip); 
}
ON_GROUPEND(Style)
{ 
	RtfFormatting& props = m_parser->getTextFormatting();
	if(props.textIsBold())
		AN_ATTRIBUTE(kAtBold, L"1");
	if(props.textIsHidden())
		AN_ATTRIBUTE(kAtHidden, L"1");
	if(props.textIsItalic())
		AN_ATTRIBUTE(kAtItalic, L"1");
	if(props.textIsStrike())
		AN_ATTRIBUTE(kAtStrike, L"1");
	if(props.textIsUnderline())
		AN_ATTRIBUTE(kAtUnderline, L"1");
	if(props.textColor() != -1 && m_parser->getOptions().doColors)
		AN_ATTRIBUTE(kAtColor, NUM_ATTR(props.textColor()));
}



ON_INITIALIZE(ListTable)
{ 
	AN_ELEMENT(kElListtable); 
}
ON_GROUPSTART(ListTable)
{ 
	AN_ANALYSER(List); 
	AN_DESTINATION(Null);
}



ON_INITIALIZE(List)
{
	AN_ELEMENT(kElListdef);
	AN_ATTRIBUTE(kAtType, kValDisc);
	AN_ATTRIBUTE(kAtOrdered, L"0");
	levelsSeen = 0;
}
ON_CONTROLWORD(List)
{
	if(cw == "listname")
		AN_DESTINATION_ATTR(kAtName);
	else if(cw == "listid" && HAS_PARAM)
		AN_ATTRIBUTE(kAtId, NUM_ATTR(param));
	
	// We let listlevel in here too
	else if(cw == "levelstartat" && HAS_PARAM)
		AN_ATTRIBUTE(kAtStart, NUM_ATTR(param));

	else if(cw == "levelnfc" && HAS_PARAM)
	{
		switch(param)
		{
        case 0: // 1, 2, 3
        case 5: // 1st, 2nd, 3rd
        case 6: // One, Two, Three
        case 7: // First, Second, Third
        case 22: // 01, 02, 03
			AN_ATTRIBUTE(kAtType, kValArabic);
            break;
        case 1: // I, II, III
			AN_ATTRIBUTE(kAtType, kValUpperRoman);
            break;
        case 2: // i, ii, iii
			AN_ATTRIBUTE(kAtType, kValLowerRoman);
            break;
        case 3: // A, B, C
			AN_ATTRIBUTE(kAtType, kValUpperAlpha);
            break;
        case 4: // a, b, c
			AN_ATTRIBUTE(kAtType, kValLowerAlpha);
            break;
        default:
			AN_ATTRIBUTE(kAtType, kValDisc);
			break;
        }

		switch(param)
		{
        case 0: case 5: case 6: case 7: case 22: 
        case 1: case 2: case 3: case 4: 
			AN_ATTRIBUTE(kAtOrdered, L"1");
			break;
		default:
			AN_ATTRIBUTE(kAtOrdered, L"0");
		}
	}

	else
		DEFAULT_CONTROLWORD;
}
ON_GROUPSTART(List)
{
	if(levelsSeen > 0)
		AN_ANALYSER(Skip);
	levelsSeen++;
}




ON_INITIALIZE(ListOverrideTable)
{
	DOM::Document document = m_parser->getDocument();
	lists = document.getElementsByTagName(kElListdef);
	curList = NULL;
	lsId = -1;
}
ON_GROUPSTART(ListOverrideTable)
{
	AN_DESTINATION(Null);
}
ON_CONTROLWORD(ListOverrideTable)
{
	// New list override clear
	if(cw == "listoverride")
		curList = NULL;

	// List id for current listoverride
	else if(cw == "listid" && HAS_PARAM)
	{
		wstring id = NUM_ATTR(param);

		if(lists != NULL)
		{
			for(int i = 0; i < lists->getLength(); i++)
			{
				DOM::Node node = lists->item(i);
				if(node != NULL && node.getNodeType() == DOM::Node::ELEMENT_NODE)
				{
					DOM::Element element = (DOM::Element&)node;
					if(element.getAttribute(kAtId) == id)
					{
						curList = element;
						break;
					}
				}
			}
		}
	}

	// The actual list code
	else if(cw == "ls" && HAS_PARAM)
		lsId = param;

	// Override the starting level for the node
	else if(cw == "levelstartat" && HAS_PARAM)
	{
		if(curList != NULL)
			curList.setAttribute(kAtStart, NUM_ATTR(param));
	}

	else
		DEFAULT_CONTROLWORD;


	// Okay before any overrides take effect we need to duplicate 
	// the list node for overriding, using the 'listid' and 'ls' we gathered
	if(curList != NULL && lsId != -1)
	{
		DOM::Element parent = (const DOM::Element&)curList.getParentNode();
		if(parent != NULL)
		{
			curList = (const DOM::Element&)curList.cloneNode(true);
			if(curList != NULL)
			{
				parent.appendChild(curList);
				curList.setAttribute(kAtList, NUM_ATTR(lsId));
			}
		}

		lsId = -1;
	}

}
ON_GROUPEND(ListOverrideTable)
{

}




ON_INITIALIZE(Info)
{
	// Create a new element
	AN_ELEMENT(kElInfo);
	AN_DESTINATION(Null);
}
ON_CONTROLWORD(Info)
{
	// The title
	if(cw == "title")
	{
		AN_ELEMENT(kElTitle);
		AN_DESTINATION(Raw);
	}
	else if(cw == "author")
	{
		AN_ELEMENT(kElAuthor);
		AN_DESTINATION(Raw);
	}
	else if(cw == "operator")
	{
		AN_ELEMENT(kElOperator);
		AN_DESTINATION(Raw);
	}
	else if(flags & kAsterisk)
		AN_ANALYSER(Skip);
	else
		DEFAULT_CONTROLWORD;
}




ON_INITIALIZE(Root)
{

}
ON_CONTROLWORD(Root)
{
	if(cw == "stylesheet")
		AN_ANALYSER(Stylesheet);
	else if(cw == "listtable")
		AN_ANALYSER(ListTable);
	else if(cw == "listoverridetable")
		AN_ANALYSER(ListOverrideTable);
	else if(cw == "info")
		AN_ANALYSER(Info);
	else if(cw == "fonttbl")
		AN_ANALYSER(Skip);
	else if(cw == "colortbl")
		AN_ANALYSER(Skip);
    else if(cw == "footnote")
        AN_ANALYSER(FootNote);
	else if(cw == "pict")
	{
		AN_ANALYSER(Skip);
		AN_DESTINATION(Null);
	}
	else if(flags & kAsterisk)
		AN_ANALYSER(Skip);
	else if(processTextContent(cw, flags, param))
		DUMMY;
    else if(processTextAutoContent(cw, flags, param))
        DUMMY;
	else if(processTextFormatting(cw, flags, param))
		DUMMY;
	else
		DEFAULT_CONTROLWORD;
}


ON_INITIALIZE(Content)
{
	parent = m_parser->getElement();
	created = false;
}
ON_CHARDATA(Content)
{
	// Create the first time we get content
	if(!created)
	{
		DOM::Element dest = m_parser->createElement(kElDest);
		parent.appendChild(dest);
		m_parser->replaceElement(dest);

		DOM::Element el = m_parser->createElement(kElBlock);
		m_parser->pushElement(el);
		m_parser->popElement();

		created = true;
	}

	if(data.length() == 0)
		return;

	int elements = 0;
	RtfFormatting& format = m_parser->getTextFormatting();

	// Now do text Properties if necessary
	if(format.textIsBold())
	{
		AN_ELEMENT(kElB);
		elements++;
	}
	if(format.textIsHidden())
	{
		AN_ELEMENT(kElHide);
		elements++;
	}
	if(format.textIsItalic())
	{
		AN_ELEMENT(kElI);
		elements++;
	}
	if(format.textIsStrike())
	{
		AN_ELEMENT(kElStrike);
		elements++;
	}
	if(format.textIsUnderline())
	{
		AN_ELEMENT(kElU);
		elements++;
	}
	if(format.textColor() != -1 && m_parser->getOptions().doColors)
	{
		AN_ELEMENT(kElColor);
		AN_ATTRIBUTE(kAtIndex, NUM_ATTR(format.textColor()));
		elements++;
	}
    if(format.textSuScript() == RtfFormatting::SUPERSCRIPT)
    {
        AN_ELEMENT(kElSuper);
        elements++;
    }
    if(format.textSuScript() == RtfFormatting::SUBSCRIPT)
    {
        AN_ELEMENT(kElSub);
        elements++;
    }

	// Write the data to the element
	m_parser->getElement().appendChild(
		m_parser->getDocument().createTextNode(data));

	// Now drop out of all the above formatting
	while(elements-- > 0)
		AN_POP_ELEMENT();
}


ON_INITIALIZE(FootNote)
{
    int ac = m_parser->getAutoCount(AUTOCOUNT_FOOTNOTE);

    AN_ELEMENT(kElFootNote);
    AN_ATTRIBUTE(kAtId, ac);
    AN_DESTINATION(Content);
}
ON_CONTROLWORD(FootNote)
{
    // Inside foot notes there's no link to the foot note
    if(cw == "chftn")
    {
        DestinationPtr dest = m_parser->getDestination();
        ASSERT(dest != NULL);
        int ac = m_parser->getAutoCount(AUTOCOUNT_FOOTNOTE);
        dest->charData(formatInt(ac));
        return;
    }
    
    // Process text content in the foot note
    else if(processTextContent(cw, flags, param))
        DUMMY;
    else if(processTextAutoContent(cw, flags, param))
        DUMMY;
    else if(processTextFormatting(cw, flags, param))
        DUMMY;
    else
        DEFAULT_CONTROLWORD;
}
ON_DONE(FootNote)
{
    m_parser->incrementAutoCount(AUTOCOUNT_FOOTNOTE);
}




ON_CHARDATA(Raw)
{
	// Write the data to the element
	m_parser->getElement().appendChild(
		m_parser->getDocument().createTextNode(data));
}




ON_INITIALIZE(Attribute)
{
	element = m_parser->getElement();
	ASSERT(element != NULL);
}

ON_CHARDATA(Attribute)
{
	// Get the current value
	wstring cur = element.getAttribute(name);

	if(data.at(data.size() - 1) == L';')
		data.resize(data.size() - 1);

	// Append data
	cur.append(data);

	// Write it back
	element.setAttribute(name, cur);
}

wstring RtfParser::formatInt(int num)
{
  char buff[16];

  // Certain OSs don't support swprintf :(
  sprintf(buff, "%d", num);

  wstring n;
  for(char* s = buff; *s; s++)
    n.append(1, *s);

  return n;
}

